"""Load dataset configurations from YAML files.

This module handles all YAML-related functionality for loading dataset
configurations, including Inspect AI-compatible `eval.yaml` files from
Hugging Face Hub repositories.
"""

import dataclasses
import logging
from pathlib import Path

import yaml
from huggingface_hub import HfApi

from .data_models import DatasetConfig, Task
from .languages import Language, get_all_languages
from .logging_utils import log_once
from .metrics.llm_as_a_judge import create_model_graded_fact_metric
from .split_utils import get_repo_splits
from .tasks import OPEN_ENDED_QA, get_all_tasks


def load_yaml_config(
    hf_api: HfApi, dataset_id: str, cache_dir: Path
) -> DatasetConfig | None:
    """Load a dataset config from an eval.yaml file in a Hugging Face repo.

    Args:
        hf_api:
            The Hugging Face API object.
        dataset_id:
            The ID of the dataset to get the config for.
        cache_dir:
            The directory to store the cache in.

    Returns:
        The dataset config if it exists, otherwise None.
    """
    external_config_path = cache_dir / "external_dataset_configs" / dataset_id
    external_config_path.mkdir(parents=True, exist_ok=True)
    hf_api.hf_hub_download(
        repo_id=dataset_id,
        repo_type="dataset",
        filename="eval.yaml",
        local_dir=external_config_path,
        local_dir_use_symlinks=False,
    )

    repo_dataset_info = hf_api.dataset_info(repo_id=dataset_id)
    fallback_language_codes: list[str] | None = None
    if repo_dataset_info.card_data is not None:
        lang_meta = getattr(repo_dataset_info.card_data, "language", None)
        if isinstance(lang_meta, list) and lang_meta:
            fallback_language_codes = [str(c) for c in lang_meta if c]

    inspect_ai_config: str | None = None
    inspect_ai_split: str | None = None
    yaml_file_path = external_config_path / "eval.yaml"
    try:
        with yaml_file_path.open(encoding="utf-8") as fh:
            raw_peek = yaml.safe_load(fh)
        if isinstance(raw_peek, dict):
            tasks_peek = raw_peek.get("tasks")
            if isinstance(tasks_peek, list) and tasks_peek:
                first_task_peek = tasks_peek[0]
                if isinstance(first_task_peek, dict):
                    config_val = first_task_peek.get("config")
                    if isinstance(config_val, str) and config_val:
                        inspect_ai_config = config_val
                    split_val = first_task_peek.get("split")
                    if isinstance(split_val, str) and split_val:
                        inspect_ai_split = split_val
    except (yaml.YAMLError, OSError):
        pass

    repo_dataset_config = load_dataset_config_from_yaml(
        yaml_path=yaml_file_path, fallback_language_codes=fallback_language_codes
    )
    if repo_dataset_config is None:
        return None

    train_split, val_split, auto_test_split = get_repo_splits(
        hf_api=hf_api, dataset_id=dataset_id
    )
    test_split = inspect_ai_split if inspect_ai_split is not None else auto_test_split
    if test_split is None:
        log_once(
            message=(
                f"Dataset {dataset_id} does not have a test split, so we cannot load "
                "it. Please ensure that the dataset has a test split."
            ),
            level=logging.ERROR,
        )
        return None

    if train_split is None and val_split is not None:
        log_once(
            message=(
                f"Dataset {dataset_id!r} has no training split. Using the validation "
                f"split {val_split!r} as the training split instead."
            ),
            level=logging.DEBUG,
        )
        train_split = val_split
        val_split = None

    source = f"{dataset_id}::{inspect_ai_config}" if inspect_ai_config else dataset_id

    repo_dataset_config.name = dataset_id
    repo_dataset_config.pretty_name = dataset_id
    repo_dataset_config.source = source
    repo_dataset_config.train_split = train_split
    repo_dataset_config.val_split = val_split
    repo_dataset_config.test_split = test_split
    return repo_dataset_config


def load_dataset_config_from_yaml(
    yaml_path: Path, fallback_language_codes: list[str] | None = None
) -> DatasetConfig | None:
    """Load a dataset config from a YAML file.

    The file is fully compatible with the Inspect AI `eval.yaml` format
    (https://inspect.aisi.org.uk/tasks.html#hugging-face). The EuroEval-specific
    `task` and `languages` keys are optional:

    * `task` -- if absent, the task is inferred from Inspect AI hints: a solver
      with `name: multiple_choice` or a `field_spec.choices` entry both map to the
      `multiple-choice` task. If the task cannot be inferred an error is logged and
      None is returned.
    * `languages` -- if absent, the `fallback_language_codes` argument (a list
      of ISO 639-1 codes) is used. When called from
      `try_get_dataset_config_from_repo`, the Hugging Face Hub repo metadata
      supplies this fallback automatically. If neither source provides a language
      list, English (`"en"`) is used as the final fallback and a warning is logged.

    Column mappings may be specified either as flat top-level keys
    (`input_column` / `target_column` / `choices_column`) or via a
    `tasks[0].field_spec` block using the Inspect AI `input` / `target` /
    `choices` sub-keys. Top-level keys take precedence when both are present.

    `tasks[0].split` is used as the test split. `try_get_dataset_config_from_repo`
    auto-detects the train and val splits from the repository, and also uses
    `tasks[0].config` as the HuggingFace dataset config/subset name.

    When reading `field_spec`:

    * `field_spec.input` is used as `input_column`.
    * `field_spec.target` is used as `target_column` only when it is a plain
      column name. Inspect AI also allows `"literal:<value>"` (a hard-coded
      answer string) and bare integers (which Inspect AI maps to letters A, B, C
      ...); both are silently skipped because they are not column names.
    * `field_spec.choices` is used as `choices_column` (a single column name or
      a list of column names).

    Example -- EuroEval flat format:

        task: classification
        languages:
          - en
        labels:
          - positive
          - negative

    Example -- pure Inspect AI format (task and languages are inferred automatically):

        # eval.yaml -- no EuroEval-specific keys required
        name: My Dataset
        tasks:
          - id: my_dataset
            split: test
            field_spec:
              input: question
              target: answer
              choices: options
            solvers:
              - name: multiple_choice
            scorers:
              - name: choice

    Example -- Inspect AI format with optional EuroEval overrides:

        # eval.yaml
        name: My Dataset
        tasks:
          - id: my_dataset
            split: test
            field_spec:
              input: text
              target: label
            solvers:
              - name: multiple_choice
            scorers:
              - name: choice
        # EuroEval-specific keys (optional; ignored by Inspect AI)
        task: multiple-choice
        languages:
          - en

    Args:
        yaml_path:
            Path to the YAML config file.
        fallback_language_codes:
            ISO 639-1 language codes to use when the YAML file does not contain a
            `languages` key. Typically supplied from HuggingFace Hub repo metadata
            by `try_get_dataset_config_from_repo`.

    Returns:
        A `DatasetConfig` built from the YAML data, or None if the file could not
            be parsed or contains invalid values.
    """
    raw = load_yaml_file(yaml_path=yaml_path)
    if raw is None:
        return None

    promote_field_spec_fields(raw=raw)

    task_obj = validate_and_get_task(raw=raw, yaml_path=yaml_path)
    if task_obj is None:
        return None

    language_objs = parse_languages(
        raw=raw, fallback_codes=fallback_language_codes, yaml_path=yaml_path
    )
    if language_objs is None:
        return None

    kwargs = build_kwargs(raw=raw, yaml_path=yaml_path)
    if kwargs is None:
        return None

    return DatasetConfig(
        task=task_obj,
        languages=language_objs,
        **kwargs,  # pyrefly: ignore[bad-argument-type]
    )


def promote_field_spec_fields(raw: dict[str, object]) -> None:
    """Promote column names from field_spec to top-level keys.

    Promotes the following mappings when the top-level key is not already set:

    * `field_spec.input` -> `input_column`
    * `field_spec.target` -> `target_column` (only if plain, not literal/int)
    * `field_spec.choices` -> `choices_column`
    * `tasks[0].split` -> `test_split`

    Args:
        raw:
            The parsed YAML data to modify in place.
    """
    tasks_raw = raw.get("tasks")
    if not isinstance(tasks_raw, list) or not tasks_raw:
        return

    first_task = tasks_raw[0]
    if not isinstance(first_task, dict):
        return

    field_spec = first_task.get("field_spec")
    if isinstance(field_spec, dict):
        if "input" in field_spec and "input_column" not in raw:
            raw["input_column"] = field_spec["input"]

        if "target" in field_spec and "target_column" not in raw:
            target = field_spec["target"]
            if isinstance(target, str) and not target.startswith("literal:"):
                raw["target_column"] = target

        if "choices" in field_spec and "choices_column" not in raw:
            raw["choices_column"] = field_spec["choices"]

    split_val = first_task.get("split")
    if isinstance(split_val, str) and split_val and "test_split" not in raw:
        raw["test_split"] = split_val


def validate_and_get_task(raw: dict[str, object], yaml_path: Path) -> Task | None:
    """Validate the task field or infer it from Inspect AI hints.

    Args:
        raw:
            The parsed YAML data.
        yaml_path:
            Path to the YAML config file (for error messages).

    Returns:
        A valid Task object, or None if validation failed.
    """
    task_map = get_all_tasks()
    task_name = raw.get("task")

    if isinstance(task_name, str):
        task_obj = task_map.get(task_name)
        if task_obj is None:
            log_once(
                message=(
                    f"Unknown task '{task_name}' in YAML config at {yaml_path}. "
                    f"Valid task names are: {sorted(task_map)}."
                ),
                level=logging.ERROR,
            )
            return None
    else:
        task_obj = infer_task_from_inspect_ai(raw=raw, task_map=task_map)
        if task_obj is None:
            log_once(
                message=(
                    f"YAML config at {yaml_path} does not contain a 'task' field and "
                    "the task could not be inferred from the Inspect AI 'tasks' block. "
                    "Add a top-level 'task' key (e.g. 'task: classification') or "
                    "include a 'multiple_choice' solver / a 'choices' field_spec entry "
                    "so that the task can be detected automatically."
                ),
                level=logging.ERROR,
            )
            return None

    return task_obj


def infer_task_from_inspect_ai(
    raw: dict[str, object], task_map: dict[str, Task]
) -> Task | None:
    """Try to infer the EuroEval task from Inspect AI YAML fields.

    Currently detects:

    * A solver with `name: multiple_choice` in `tasks[0].solvers`
      -> `multiple-choice`
    * A `choices` key in `tasks[0].field_spec` -> `multiple-choice`
    * A scorer with `name: model_graded_fact` in `tasks[0].scorers`
      -> `open-ended-qa` task with an LLM-as-a-judge metric.
      The judge model is read from `scorers[0].args.model`; when absent, the
      default judge defined in `OPEN_ENDED_QA` is used.

    Args:
        raw:
            The raw YAML data.
        task_map:
            The mapping from task names to task objects.

    Returns:
        The inferred task, or None if the task cannot be inferred.
    """
    tasks_raw = raw.get("tasks")
    if not isinstance(tasks_raw, list) or not tasks_raw:
        return None
    first_task = tasks_raw[0]
    if not isinstance(first_task, dict):
        return None

    solvers = first_task.get("solvers")
    if isinstance(solvers, list):
        for solver in solvers:
            if isinstance(solver, dict) and solver.get("name") == "multiple_choice":
                return task_map.get("multiple-choice")

    scorers = first_task.get("scorers")
    if isinstance(scorers, list):
        for scorer in scorers:
            if isinstance(scorer, dict) and scorer.get("name") == "model_graded_fact":
                judge_id: str | None = None
                args = scorer.get("args")
                if isinstance(args, dict):
                    model_val = args.get("model")
                    if isinstance(model_val, str) and model_val:
                        judge_id = model_val
                if judge_id is not None:
                    metric = create_model_graded_fact_metric(judge_id=judge_id)
                    return dataclasses.replace(OPEN_ENDED_QA, metrics=[metric])
                return OPEN_ENDED_QA

    field_spec = first_task.get("field_spec")
    if isinstance(field_spec, dict) and "choices" in field_spec:
        return task_map.get("multiple-choice")

    return None


def parse_languages(
    raw: dict[str, object], fallback_codes: list[str] | None, yaml_path: Path
) -> list[Language] | None:
    """Parse language codes from YAML or use fallbacks.

    Args:
        raw:
            The parsed YAML data.
        fallback_codes:
            ISO 639-1 language codes to use as a fallback.
        yaml_path:
            Path to the YAML config file (for error messages).

    Returns:
        A list of Language objects, or None if validation failed.
    """
    language_map = get_all_languages()
    raw_languages = raw.get("languages")

    if isinstance(raw_languages, list) and raw_languages:
        language_codes: list[str] = [str(c) for c in raw_languages]
    elif fallback_codes:
        log_once(
            message=(
                f"YAML config at {yaml_path} does not contain a 'languages' key. "
                "Using language(s) from the repository metadata: "
                f"{fallback_codes}."
            ),
            level=logging.DEBUG,
        )
        language_codes = fallback_codes
    else:
        log_once(
            message=(
                f"YAML config at {yaml_path} does not contain a 'languages' key and "
                "no language metadata could be found for this repository. Defaulting "
                "to English. Add a top-level 'languages' key to the YAML file "
                "(e.g. 'languages: [en]') to override this."
            ),
            level=logging.WARNING,
        )
        language_codes = ["en"]

    language_objs: list[Language] = []
    for code in language_codes:
        lang = language_map.get(code)
        if lang is None:
            log_once(
                message=(
                    f"Unknown language code '{code}' in YAML config at {yaml_path}."
                ),
                level=logging.ERROR,
            )
            return None
        language_objs.append(lang)

    return language_objs


def build_kwargs(
    raw: dict[str, object], yaml_path: Path
) -> dict[str, str | int | list[str] | dict[str, str]] | None:
    """Build keyword arguments for `DatasetConfig` from YAML fields.

    Reads the following optional fields from `raw` and maps them to the
    corresponding `DatasetConfig` constructor arguments:

    * String fields: `prompt_prefix`, `prompt_template`, `instruction_prompt`,
      `input_column`, `target_column`, `test_split`.
    * Integer fields: `num_few_shot_examples`, `max_generated_tokens`.
    * `labels` -- a list of strings.
    * `prompt_label_mapping` -- a mapping from strings to strings.
    * `choices_column` -- a string or list of strings.

    Args:
        raw:
            The parsed YAML data.
        yaml_path:
            Path to the YAML config file (for error messages).

    Returns:
        A dictionary suitable for unpacking into `DatasetConfig(...)`, or None
            if any field fails validation.
    """
    kwargs: dict[str, str | int | list[str] | dict[str, str]] = {}

    for field_name in (
        "prompt_prefix",
        "prompt_template",
        "instruction_prompt",
        "input_column",
        "target_column",
        "test_split",
    ):
        value = parse_string_field(raw=raw, field_name=field_name, yaml_path=yaml_path)
        if value is not None:
            kwargs[field_name] = value

    for field_name in ("num_few_shot_examples", "max_generated_tokens"):
        value = parse_int_field(raw=raw, field_name=field_name, yaml_path=yaml_path)
        if value is not None:
            kwargs[field_name] = value

    labels_raw = raw.get("labels")
    if labels_raw is not None:
        if not isinstance(labels_raw, list):
            log_once(
                message=f"Field 'labels' in YAML config at {yaml_path} must be a list.",
                level=logging.ERROR,
            )
            return None
        kwargs["labels"] = [str(lbl) for lbl in labels_raw]

    prompt_label_mapping_raw = raw.get("prompt_label_mapping")
    if prompt_label_mapping_raw is not None:
        if not isinstance(prompt_label_mapping_raw, dict):
            log_once(
                message=(
                    "Field 'prompt_label_mapping' in YAML config at"
                    f" {yaml_path} must be a mapping."
                ),
                level=logging.ERROR,
            )
            return None
        kwargs["prompt_label_mapping"] = {
            str(k): str(v) for k, v in prompt_label_mapping_raw.items()
        }

    choices_column_raw = raw.get("choices_column")
    if choices_column_raw is not None:
        if isinstance(choices_column_raw, list):
            kwargs["choices_column"] = [str(c) for c in choices_column_raw]
        elif isinstance(choices_column_raw, str):
            kwargs["choices_column"] = choices_column_raw
        else:
            log_once(
                message=(
                    "Field 'choices_column' in YAML config at"
                    f" {yaml_path} must be a string or a list of strings."
                ),
                level=logging.ERROR,
            )
            return None

    return kwargs


def parse_string_field(
    raw: dict[str, object], field_name: str, yaml_path: Path
) -> str | None:
    """Parse and validate a string field from YAML.

    Args:
        raw:
            The parsed YAML data.
        field_name:
            The name of the field to parse.
        yaml_path:
            Path to the YAML config file (for error messages).

    Returns:
        The field value as a string, or None if validation failed.
    """
    value = raw.get(field_name)
    if value is not None:
        if not isinstance(value, str):
            log_once(
                message=(
                    f"Field '{field_name}' in YAML config at {yaml_path} must be a "
                    "string."
                ),
                level=logging.ERROR,
            )
            return None
        return value
    return None


def parse_int_field(
    raw: dict[str, object], field_name: str, yaml_path: Path
) -> int | None:
    """Parse and validate an integer field from YAML.

    Args:
        raw:
            The parsed YAML data.
        field_name:
            The name of the field to parse.
        yaml_path:
            Path to the YAML config file (for error messages).

    Returns:
        The field value as an integer, or None if validation failed.
    """
    value = raw.get(field_name)
    if value is not None:
        if isinstance(value, bool) or not isinstance(value, int):
            log_once(
                message=(
                    f"Field '{field_name}' in YAML config at {yaml_path} must be an "
                    "integer."
                ),
                level=logging.ERROR,
            )
            return None
        return value
    return None


def load_yaml_file(yaml_path: Path) -> dict[str, object] | None:
    """Load a YAML file and return its contents as a dictionary.

    Args:
        yaml_path:
            Path to the YAML config file.

    Returns:
        The parsed YAML content as a dictionary, or None if parsing failed.
    """
    try:
        with yaml_path.open(encoding="utf-8") as fh:
            raw = yaml.safe_load(fh)
    except yaml.YAMLError as exc:
        log_once(
            message=f"Could not parse YAML config from {yaml_path}: {exc}",
            level=logging.ERROR,
        )
        return None

    if not isinstance(raw, dict):
        log_once(
            message=f"YAML config at {yaml_path} must be a mapping at the top level.",
            level=logging.ERROR,
        )
        return None

    return raw



# /// script
# requires-python = ">=3.10,<4.0"
# dependencies = [
#     "datasets==3.5.0",
#     "huggingface-hub==0.24.0",
#     "pandas==2.2.0",
#     "requests==2.32.3",
#     "tqdm==4.67.1",
# ]
# ///

"""Create the NorNE-mini NER datasets and upload them to the HF Hub."""

import re
import typing as t
from collections import defaultdict

import pandas as pd
import requests
from datasets import Split
from datasets.arrow_dataset import Dataset
from datasets.dataset_dict import DatasetDict
from huggingface_hub.hf_api import HfApi
from tqdm.auto import tqdm


def main() -> None:
    """Create the NorNE-mini NER datasets and uploads them to the HF Hub."""
    # Define the base download URLs
    base_urls = dict(
        nb="https://github.com/UniversalDependencies/UD_Norwegian-Bokmaal/raw/master/"
        "no_bokmaal-ud-{}.conllu",
        nn="https://github.com/UniversalDependencies/UD_Norwegian-Nynorsk/raw/master/"
        "no_nynorsk-ud-{}.conllu",
    )

    # Define conversion dict to be used with the NER tags
    ner_conversion_dict = {
        "O": "O",
        "B-LOC": "B-LOC",
        "I-LOC": "I-LOC",
        "B-PER": "B-PER",
        "I-PER": "I-PER",
        "B-ORG": "B-ORG",
        "I-ORG": "I-ORG",
        "B-MISC": "B-MISC",
        "I-MISC": "I-MISC",
        "B-GPE_LOC": "B-LOC",
        "I-GPE_LOC": "I-LOC",
        "B-GPE_ORG": "B-ORG",
        "I-GPE_ORG": "I-ORG",
        "B-PROD": "B-MISC",
        "I-PROD": "I-MISC",
        "B-DRV": "B-MISC",
        "I-DRV": "I-MISC",
        "B-EVT": "B-MISC",
        "I-EVT": "I-MISC",
    }

    # Set up the progress bar and iterate over the languages
    with tqdm(["nb", "nn"], desc="Creating NorNE-mini datasets") as pbar:
        for lang in pbar:
            # Define download URLs
            train_url = base_urls[lang].format("train")
            val_url = base_urls[lang].format("dev")
            test_url = base_urls[lang].format("test")

            # Download the data
            data = dict(
                train=requests.get(train_url).text.split("\n"),
                val=requests.get(val_url).text.split("\n"),
                test=requests.get(test_url).text.split("\n"),
            )

            # Update the progress bar description
            pbar.set_description(f"Creating NorNE datasets - {lang}")

            # Iterate over the data splits
            dfs = dict()
            for split, lines in data.items():
                # Initialise the records, data dictionary and document
                records = list()
                data_dict: t.Dict[str, t.List[t.Union[int, str]]] = defaultdict(list)
                doc = ""

                # Iterate over the data for the given split
                for line in lines:
                    # If we are at the first line of an entry then extract the document
                    if line.startswith("# text = "):
                        doc = re.sub("# text = ", "", line)

                    # Otherwise, if the line is a comment then ignore it
                    elif line.startswith("#"):
                        continue

                    # Otherwise, if we have reached the end of an entry then store it
                    # to the list of records and reset the data dictionary and document
                    elif line == "":
                        merged_data_dict: t.Dict[
                            str, t.Union[str, t.List[t.Union[int, str]]]
                        ]
                        if len(data_dict["tokens"]) > 0:
                            merged_data_dict = {**data_dict, "doc": doc}
                            records.append(merged_data_dict)
                        data_dict = defaultdict(list)
                        doc = ""

                    # Otherwise we are in the middle of an entry which is not a
                    # comment, so we extract the data from the line and store it in the
                    # data dictionary
                    else:
                        data_tup = line.split("\t")
                        ner_tag = data_tup[9].replace("name=", "").split("|")[-1]
                        data_dict["ids"].append(data_tup[0])
                        data_dict["tokens"].append(data_tup[1])
                        data_dict["ner_tags"].append(ner_conversion_dict[ner_tag])

                # Convert the records to a dataframe and store it
                dfs[split] = pd.DataFrame.from_records(records)

            # Merge the dataframes
            df = pd.concat(dfs.values()).reset_index(drop=True)

            # Remove duplicates
            df = df.drop_duplicates().reset_index(drop=True)

            # Create new splits
            val_df = df.sample(n=256, random_state=4242)
            df_filtered = df[~df.index.isin(val_df.index)]
            test_df = df_filtered.sample(n=2048, random_state=4242)
            full_train_df = df_filtered[~df_filtered.index.isin(test_df.index)]
            assert isinstance(full_train_df, pd.DataFrame)
            train_df = full_train_df.sample(n=1024, random_state=4242)

            assert isinstance(train_df, pd.DataFrame)
            assert isinstance(val_df, pd.DataFrame)
            assert isinstance(test_df, pd.DataFrame)

            # Collect datasets in a dataset dictionary
            dataset = DatasetDict(
                {
                    "train": Dataset.from_pandas(train_df, split=Split.TRAIN),
                    "val": Dataset.from_pandas(val_df, split=Split.VALIDATION),
                    "test": Dataset.from_pandas(test_df, split=Split.TEST),
                    "full_train": Dataset.from_pandas(
                        full_train_df,
                        split="full_train",  # type: ignore[bad-argument-type]
                    ),
                }
            )

            # Push the dataset to the Hugging Face Hub
            dataset_id = f"EuroEval/norne-{lang}-mini"
            HfApi().delete_repo(dataset_id, repo_type="dataset", missing_ok=True)
            dataset.push_to_hub(dataset_id, private=True)


if __name__ == "__main__":
    main()



# /// script
# requires-python = ">=3.10,<4.0"
# dependencies = [
#     "datasets==3.5.0",
#     "huggingface-hub==0.24.0",
#     "openai==1.66.5",
#     "pandas==2.2.0",
#     "pydantic==2.6.0",
#     "python-dotenv==1.0.1",
#     "requests==2.32.3",
#     "tqdm==4.67.1",
# ]
# ///

"""Create the MultiNRC datasets and upload them to the HF Hub."""

import json
import logging
import os
import random
import re

import pandas as pd
from constants import CHOICES_MAPPING
from datasets import Dataset, DatasetDict, Split, load_dataset
from dotenv import load_dotenv
from huggingface_hub import HfApi
from openai import OpenAI
from openai.types.chat import ChatCompletionUserMessageParam
from pydantic import BaseModel
from tqdm.auto import tqdm

logging.basicConfig(format="%(asctime)s ⋅ %(message)s", level=logging.INFO)
logger = logging.getLogger("create_multinrc")

load_dotenv()


class CandidateAnswers(BaseModel):
    """Candidate answers from the OpenAI API."""

    first: str
    second: str
    third: str


LABELS = ["a", "b", "c", "d"]

LANGUAGE_SUBSET_MAPPING = {"es": "spanish", "fr": "french"}


def main() -> None:
    """Create the MultiNRC datasets and upload them to the HF Hub."""
    repo_id = "ScaleAI/MultiNRC"

    # The dataset has a single subset; load it once and filter per language
    full_dataset = load_dataset(path=repo_id, split="test", token=True)
    assert isinstance(full_dataset, Dataset)

    for language in LANGUAGE_SUBSET_MAPPING.keys():
        language_name = LANGUAGE_SUBSET_MAPPING[language].capitalize()

        # Filter by the 'language' column (values are e.g. 'Spanish', 'French')
        dataset = full_dataset.filter(
            lambda row, lang=language_name: row["language"] == lang
        )
        assert isinstance(dataset, Dataset)

        # Build the multiple choice dataset using a language model
        df = build_dataset_with_llm(dataset=dataset, language=language)

        # Create splits: 64 train, 128 val, rest test
        train_df = df.sample(64, random_state=4242)
        df = df.drop(train_df.index.tolist())

        val_df = df.sample(128, random_state=4242)
        test_df = df.drop(val_df.index.tolist())

        # Reset the index
        train_df = train_df.reset_index(drop=True)
        val_df = val_df.reset_index(drop=True)
        test_df = test_df.reset_index(drop=True)

        # Collect datasets in a dataset dictionary
        dataset = DatasetDict(
            {
                "train": Dataset.from_pandas(train_df, split=Split.TRAIN),
                "val": Dataset.from_pandas(val_df, split=Split.VALIDATION),
                "test": Dataset.from_pandas(test_df, split=Split.TEST),
            }
        )

        # Create dataset ID
        dataset_id = f"EuroEval/multinrc-{language}"

        # Remove the dataset from Hugging Face Hub if it already exists
        HfApi().delete_repo(dataset_id, repo_type="dataset", missing_ok=True)

        # Push the dataset to the Hugging Face Hub
        dataset.push_to_hub(dataset_id, private=True)


def build_dataset_with_llm(dataset: Dataset, language: str) -> pd.DataFrame:
    """Build the multiple choice dataset using a language model.

    Args:
        dataset:
            The dataset containing the questions and correct answers.
        language:
            The ISO language code (e.g. "es", "fr").

    Returns:
        The multiple choice dataset as a DataFrame with 'text' and 'label' columns.
    """
    df = dataset.to_pandas()
    assert isinstance(df, pd.DataFrame)

    client = OpenAI(api_key=os.environ["OPENAI_API_KEY"])

    cache_file = f"multinrc_{language}_cache.json"
    if os.path.exists(cache_file):
        with open(cache_file) as f:
            cache = json.load(f)
    else:
        cache = {}

    language_name = LANGUAGE_SUBSET_MAPPING[language].capitalize()
    texts: list[str] = []
    correct_labels: list[str] = []
    df_len = len(df)

    for i, row in tqdm(
        df.iterrows(), total=df_len, desc=f"Computing LLM responses ({language})"
    ):
        id_ = str(i)

        if id_ not in cache:
            logger.info(f"Processing id: {id_}/{df_len}")
            messages: list[ChatCompletionUserMessageParam] = []

            user_message = ChatCompletionUserMessageParam(
                role="user",
                content=(
                    f"For the following {language_name} question: "
                    f"'{row.i18n_prompt}' where the correct answer is: "
                    f"'{row.i18n_gtfa}', please provide 3 plausible but incorrect "
                    f"alternative answers in {language_name}. Return the alternatives "
                    "in a JSON dictionary with keys 'first', 'second', and 'third'. "
                    "The values should be the alternatives only, without any numbering "
                    "or formatting. The alternatives should be unique and must not "
                    "contain or repeat the correct answer."
                ),
            )
            messages.append(user_message)

            completion = client.beta.chat.completions.parse(
                model="gpt-4.1", messages=messages, response_format=CandidateAnswers
            )

            event = completion.choices[0].message.parsed
            assert event is not None, f"Expected a response, but got {event}."
            cache[id_] = dict(event)
            with open(cache_file, "w") as f:
                json.dump(cache, f)

        options_raw = cache[id_]
        labels = LABELS.copy()
        random.shuffle(labels)

        correct_answer = str(row.i18n_gtfa)
        options = {
            labels[0]: re.sub(r"^[0-9]\. *", "", options_raw["first"]),
            labels[1]: re.sub(r"^[0-9]\. *", "", options_raw["second"]),
            labels[2]: re.sub(r"^[0-9]\. *", "", options_raw["third"]),
            labels[3]: correct_answer,
        }
        options = dict(sorted(options.items()))

        if len(set(options.values())) != 4:
            logger.warning(
                f"The options are not unique for question '{row.i18n_prompt}', got "
                f"{options}. Skipping."
            )
            continue

        correct_label = [k for k, v in options.items() if v == correct_answer][0]

        text = (
            str(row.i18n_prompt).replace("\n", " ").strip()
            + f"\n{CHOICES_MAPPING[language]}:\n"
            + f"a. {options['a']}\n"
            + f"b. {options['b']}\n"
            + f"c. {options['c']}\n"
            + f"d. {options['d']}"
        )

        # Sanity check that the choices are at the end of the text
        sections = text.split("\n")
        choice_idxs = [
            idx
            for idx, section in enumerate(sections)
            if re.match(pattern=r"^[a-d]\. ", string=section) is not None
        ]
        if not all(
            choice_idx == len(sections) - i
            for i, choice_idx in enumerate(sorted(choice_idxs, reverse=True), start=1)
        ):
            logger.warning(
                f"Choices are not at the end of the document for '{text}'. Skipping."
            )
            continue

        texts.append(text)
        correct_labels.append(correct_label)

    return pd.DataFrame({"text": texts, "label": correct_labels})


if __name__ == "__main__":
    main()



# /// script
# requires-python = ">=3.10,<4.0"
# dependencies = [
#     "datasets==3.5.0",
#     "huggingface-hub==0.24.0",
#     "pandas==2.2.0",
#     "requests==2.32.3",
#     "scikit-learn<1.6.0",
# ]
# ///

"""Create the Danish Sentiment in Context dataset and upload it to the HF Hub."""

import io
from zipfile import ZipFile

import pandas as pd
import requests as rq
from datasets import Dataset, DatasetDict, Split
from huggingface_hub import HfApi
from sklearn.model_selection import train_test_split


def main() -> None:
    """Create the Danish Sentiment in Context dataset and upload it to the HF Hub."""
    # Download the ZIP file from the Danish Semantic Reasoning Benchmark repository
    url = (
        "https://github.com/kuhumcst/danish-semantic-reasoning-benchmark/raw/main"
        "/sentiment/sentiment.zip"
    )
    response = rq.get(url=url)
    response.raise_for_status()

    # Extract the TSV file from the ZIP
    with ZipFile(file=io.BytesIO(initial_bytes=response.content)) as zip_file:
        tsv_files = [
            zip_file.read(name=file_name, pwd=b"benchmark")
            for file_name in zip_file.namelist()
            if file_name.endswith(".tsv")
        ]
        assert len(tsv_files) == 1, (
            f"Expected one TSV file in the ZIP file, but found {len(tsv_files)}."
        )
        df = pd.read_csv(
            filepath_or_buffer=io.BytesIO(initial_bytes=tsv_files[0]), sep="\t"
        )

    # Build the 'text' column by combining the target word and context
    df["text"] = "Ord: " + df["target"] + "\nKontekst: " + df["context"]

    # Map the sentiment scale (-3 to +3) to positive/neutral/negative labels
    def map_label(score: int) -> str:
        if score < 0:
            return "negative"
        elif score > 0:
            return "positive"
        else:
            return "neutral"

    df["label"] = df["label"].apply(map_label)

    # Keep only the relevant columns
    df = df[["text", "label"]].copy()

    # Remove duplicates
    df.drop_duplicates(inplace=True)
    df.reset_index(drop=True, inplace=True)

    # Create validation split
    val_size = 64
    traintest_df, val_df = train_test_split(
        df, test_size=val_size, random_state=4242, stratify=df["label"]
    )

    # Create train and test splits
    train_size = 256
    test_size = len(traintest_df) - train_size
    train_df, test_df = train_test_split(
        traintest_df,
        train_size=train_size,
        test_size=test_size,
        random_state=4242,
        stratify=traintest_df["label"],
    )

    # Reset the indices
    train_df = train_df.reset_index(drop=True)
    val_df = val_df.reset_index(drop=True)
    test_df = test_df.reset_index(drop=True)

    # Collect datasets in a dataset dictionary
    dataset = DatasetDict(
        {
            "train": Dataset.from_pandas(train_df, split=Split.TRAIN),
            "val": Dataset.from_pandas(val_df, split=Split.VALIDATION),
            "test": Dataset.from_pandas(test_df, split=Split.TEST),
        }
    )

    # Create dataset ID
    dataset_id = "EuroEval/danish-sentiment-in-context"

    # Remove the dataset from Hugging Face Hub if it already exists
    HfApi().delete_repo(dataset_id, repo_type="dataset", missing_ok=True)

    # Push the dataset to the Hugging Face Hub
    dataset.push_to_hub(dataset_id, private=True)


if __name__ == "__main__":
    main()



"""Create the BFCL-v2 dataset and upload it to the HF Hub."""

import json
import os
import random
from pathlib import Path
from urllib.request import urlopen

from datasets import Dataset, DatasetDict
from dotenv import load_dotenv


def main() -> None:
    """Create BFCL-v2 dataset (the part used by EuroEval) and push to hugging face."""
    items: list = []
    for subset_name in [
        "live_multiple",
        "live_parallel_multiple",
        "live_parallel",
        "live_simple",
        "multiple",
        "parallel_multiple",
        "parallel",
        "simple_java",
        "simple_javascript",
        "simple_python",
    ]:
        url_prefix = (
            "https://raw.githubusercontent.com/ShishirPatil/gorilla"
            "/refs/heads/main/berkeley-function-call-leaderboard/bfcl_eval/data"
        )
        input_url = f"{url_prefix}/BFCL_v4_{subset_name}.json"
        ground_truth_url = f"{url_prefix}/possible_answer/BFCL_v4_{subset_name}.json"
        print(f"Loading dataset '{subset_name}'")
        inputs = _load_jsonl_from_url(input_url)
        ground_truth = _load_jsonl_from_url(ground_truth_url)

        # Join input and ground_truth by 'id' key
        ground_truth_by_id = {item["id"]: item for item in ground_truth}
        for item in inputs:
            item_id = item["id"]
            gt: dict = ground_truth_by_id.get(item_id, {})
            joined: dict = item | gt
            items.append(joined)

    for item in items:
        function_str = json.dumps(item["function"])
        question_str = item["question"][0][0]["content"]
        item: dict
        item["text"] = f"Functions:\n{function_str}\nQuestion: {question_str}"
        item["function"] = function_str
        item["question"] = question_str
        item["target_text"] = json.dumps(item.pop("ground_truth"))

    load_dotenv()
    dataset_dict = _split_dataset_to_dict(items)
    dataset_dict.push_to_hub(
        "EuroEval/bfcl-v2", private=True, token=os.getenv("HF_TOKEN")
    )


def _split_dataset_to_dict(items: list[dict]) -> DatasetDict:
    """Split dataset into train/validation/test with deterministic random sampling.

    Args:
        items:
            List of dataset items to split

    Returns:
        DatasetDict with train, validation, and test splits
    """
    random.seed(42)
    shuffled_items = items.copy()
    random.shuffle(shuffled_items)

    train_size = 250
    val_size = 250
    # test_size = 2001

    train_items = shuffled_items[:train_size]
    val_items = shuffled_items[train_size : train_size + val_size]
    test_items = shuffled_items[train_size + val_size :]

    return DatasetDict(
        {
            "train": Dataset.from_list(train_items),
            "val": Dataset.from_list(val_items),
            "test": Dataset.from_list(test_items),
        }
    )


def _load_jsonl(path: str | Path) -> list:
    """Load jsonl.

    Args:
        path: string with serialized JSONL or Path to JSONL file

    Returns:
        List of deserialized objects
    """
    if isinstance(path, Path):
        path = path.read_text()
    return [json.loads(line) for line in path.splitlines()]


def _load_jsonl_from_url(url: str) -> list:
    """Load jsonl from url.

    Args:
        url: url to JSONL

    Returns:
        List of deserialized objects
    """
    with urlopen(url) as r:
        return _load_jsonl(r.read().decode())


if __name__ == "__main__":
    main()



"""Templates for the Word in Context task."""

import typing as t

from ..data_models import PromptConfig
from ..languages import (
    ALBANIAN,
    BELARUSIAN,
    BULGARIAN,
    CATALAN,
    CROATIAN,
    CZECH,
    DANISH,
    DUTCH,
    ENGLISH,
    ESTONIAN,
    FAROESE,
    FINNISH,
    FRENCH,
    GERMAN,
    GREEK,
    HUNGARIAN,
    ICELANDIC,
    ITALIAN,
    LATVIAN,
    LITHUANIAN,
    NORWEGIAN,
    NORWEGIAN_BOKMÅL,
    NORWEGIAN_NYNORSK,
    POLISH,
    PORTUGUESE,
    ROMANIAN,
    SERBIAN,
    SLOVAK,
    SLOVENE,
    SPANISH,
    SWEDISH,
    UKRAINIAN,
)

if t.TYPE_CHECKING:
    from ..languages import Language

WIC_TEMPLATES: dict["Language", PromptConfig] = {
    ALBANIAN: PromptConfig(
        default_prompt_label_mapping=dict(same_sense="po", different_sense="jo"),
        default_prompt_prefix="Më poshtë janë shembuj të fjalëve të përdorura në dy "
        "kontekste dhe nëse kanë të njëjtën kuptim.",
        default_prompt_template="{text}\nE njëjta kuptim: {label}",
        default_instruction_prompt="{text}\n\nA ka fjala të njëjtën kuptim në të dy "
        "kontekstet? Përgjigjuni me {labels_str}, dhe asgjë tjetër.",
    ),
    BELARUSIAN: PromptConfig(
        default_prompt_label_mapping=dict(same_sense="так", different_sense="не"),
        default_prompt_prefix="Ніжэй прыведзены прыклады слоў, якія выкарыстоўваюцца "
        "ў двух кантэкстах, і ці маюць яны аднолькавае значэнне.",
        default_prompt_template="{text}\nАднолькавае значэнне: {label}",
        default_instruction_prompt="{text}\n\nЦі мае слова аднолькавае значэнне ў "
        "абодвух кантэкстах? Адкажыце толькі {labels_str}, і нічога іншага.",
    ),
    BULGARIAN: PromptConfig(
        default_prompt_label_mapping=dict(same_sense="да", different_sense="не"),
        default_prompt_prefix="Следват примери за думи, използвани в два контекста, и "
        "дали имат едно и също значение.",
        default_prompt_template="{text}\nЕдно и също значение: {label}",
        default_instruction_prompt="{text}\n\nИма ли думата едно и също значение в "
        "двата контекста? Отговорете с {labels_str}, и нищо друго.",
    ),
    CATALAN: PromptConfig(
        default_prompt_label_mapping=dict(same_sense="sí", different_sense="no"),
        default_prompt_prefix="A continuació es mostren exemples de paraules usades en "
        "dos contextos i si tenen el mateix significat.",
        default_prompt_template="{text}\nMateix significat: {label}",
        default_instruction_prompt="{text}\n\nTé la paraula el mateix significat en "
        "els dos contextos? Respon amb {labels_str}, i res més.",
    ),
    CROATIAN: PromptConfig(
        default_prompt_label_mapping=dict(same_sense="da", different_sense="ne"),
        default_prompt_prefix="Sljedeći su primjeri riječi korištenih u dva konteksta "
        "i imaju li isto značenje.",
        default_prompt_template="{text}\nIsto značenje: {label}",
        default_instruction_prompt="{text}\n\nIma li riječ isto značenje u oba "
        "konteksta? Odgovorite s {labels_str}, i ništa drugo.",
    ),
    CZECH: PromptConfig(
        default_prompt_label_mapping=dict(same_sense="ano", different_sense="ne"),
        default_prompt_prefix="Následující jsou příklady slov použitých ve dvou "
        "kontextech a zda mají stejný význam.",
        default_prompt_template="{text}\nStejný význam: {label}",
        default_instruction_prompt="{text}\n\nMá slovo ve dvou kontextech stejný "
        "význam? Odpovězte {labels_str}, a nic jiné.",
    ),
    DANISH: PromptConfig(
        default_prompt_label_mapping=dict(same_sense="ja", different_sense="nej"),
        default_prompt_prefix="Følgende er eksempler på ord brugt i to kontekster og "
        "om de har samme betydning.",
        default_prompt_template="{text}\nSamme betydning: {label}",
        default_instruction_prompt="{text}\n\nHar ordet den samme betydning i de to "
        "kontekster? Svar kun med {labels_str}, og intet andet.",
    ),
    DUTCH: PromptConfig(
        default_prompt_label_mapping=dict(same_sense="ja", different_sense="nee"),
        default_prompt_prefix="Hieronder staan voorbeelden van woorden in twee "
        "contexten en of ze dezelfde betekenis hebben.",
        default_prompt_template="{text}\nZelfde betekenis: {label}",
        default_instruction_prompt="{text}\n\nHeeft het woord dezelfde betekenis in "
        "beide contexten? Antwoord met {labels_str}, en verder niets.",
    ),
    ENGLISH: PromptConfig(
        default_prompt_label_mapping=dict(same_sense="yes", different_sense="no"),
        default_prompt_prefix="The following are examples of words used in two "
        "contexts and whether they have the same meaning.",
        default_prompt_template="{text}\nSame meaning: {label}",
        default_instruction_prompt="{text}\n\nDoes the word have the same meaning in "
        "both contexts? Answer with {labels_str}, and nothing else.",
    ),
    ESTONIAN: PromptConfig(
        default_prompt_label_mapping=dict(same_sense="jah", different_sense="ei"),
        default_prompt_prefix="Järgnevad on näited sõnadest, mida kasutatakse kahes "
        "kontekstis, ja kas neil on sama tähendus.",
        default_prompt_template="{text}\nSama tähendus: {label}",
        default_instruction_prompt="{text}\n\nKas sõnal on mõlemas kontekstis sama "
        "tähendus? Vasta {labels_str}, ja mitte midagi muud.",
    ),
    FAROESE: PromptConfig(
        default_prompt_label_mapping=dict(same_sense="ja", different_sense="nei"),
        default_prompt_prefix="Hér eru dæmi um orð brúkt í tveimum samanhangum og um "
        "tey hava somu týdningina.",
        default_prompt_template="{text}\nSama týdningur: {label}",
        default_instruction_prompt="{text}\n\nHevur orðið somu týdningina í báðum "
        "samanhingunum? Svara við {labels_str}, og einki annað.",
    ),
    FINNISH: PromptConfig(
        default_prompt_label_mapping=dict(same_sense="kyllä", different_sense="ei"),
        default_prompt_prefix="Seuraavat ovat esimerkkejä sanoista kahdessa "
        "asiayhteydessä ja onko niillä sama merkitys.",
        default_prompt_template="{text}\nSama merkitys: {label}",
        default_instruction_prompt="{text}\n\nOnko sanalla sama merkitys molemmissa "
        "asiayhteyksissä? Vastaa {labels_str}, ja ei mitään muuta.",
    ),
    FRENCH: PromptConfig(
        default_prompt_label_mapping=dict(same_sense="oui", different_sense="non"),
        default_prompt_prefix="Les exemples suivants montrent des mots utilisés dans "
        "deux contextes et s'ils ont le même sens.",
        default_prompt_template="{text}\nMême sens : {label}",
        default_instruction_prompt="{text}\n\nLe mot a-t-il le même sens dans les deux "
        "contextes ? Répondez par {labels_str}, et rien d'autre.",
    ),
    GERMAN: PromptConfig(
        default_prompt_label_mapping=dict(same_sense="ja", different_sense="nein"),
        default_prompt_prefix="Im Folgenden sind Beispiele für Wörter in zwei "
        "Kontexten und ob sie die gleiche Bedeutung haben.",
        default_prompt_template="{text}\nGleiche Bedeutung: {label}",
        default_instruction_prompt="{text}\n\nHat das Wort in beiden Kontexten die "
        "gleiche Bedeutung? Antworten Sie mit {labels_str}, und nichts anderes.",
    ),
    GREEK: PromptConfig(
        default_prompt_label_mapping=dict(same_sense="ναι", different_sense="όχι"),
        default_prompt_prefix="Τα παρακάτω είναι παραδείγματα λέξεων που "
        "χρησιμοποιούνται σε δύο περιβάλλοντα και αν έχουν την ίδια σημασία.",
        default_prompt_template="{text}\nΊδια σημασία: {label}",
        default_instruction_prompt="{text}\n\nΈχει η λέξη την ίδια σημασία και στα δύο "
        "περιβάλλοντα; Απαντήστε με {labels_str}, και τίποτα άλλο.",
    ),
    HUNGARIAN: PromptConfig(
        default_prompt_label_mapping=dict(same_sense="igen", different_sense="nem"),
        default_prompt_prefix="A következők szavak két kontextusban való használatára "
        "és azonos jelentésükre mutatnak példát.",
        default_prompt_template="{text}\nAzonos jelentés: {label}",
        default_instruction_prompt="{text}\n\nAzonos jelentéssel bír-e a szó mindkét "
        "kontextusban? Csak {labels_str}-val válaszoljon, és semmi mással.",
    ),
    ICELANDIC: PromptConfig(
        default_prompt_label_mapping=dict(same_sense="já", different_sense="nei"),
        default_prompt_prefix="Hér fyrir neðan eru dæmi um orð notuð í tveimur "
        "samhengi og hvort þau hafi sömu merkingu.",
        default_prompt_template="{text}\nSama merking: {label}",
        default_instruction_prompt="{text}\n\nHefur orðið sömu merkingu í báðum "
        "samhengjum? Svaraðu með {labels_str}, og ekkert annað.",
    ),
    ITALIAN: PromptConfig(
        default_prompt_label_mapping=dict(same_sense="sì", different_sense="no"),
        default_prompt_prefix="Di seguito sono riportati esempi di parole usate in due "
        "contesti e se hanno lo stesso significato.",
        default_prompt_template="{text}\nStesso significato: {label}",
        default_instruction_prompt="{text}\n\nLa parola ha lo stesso significato in "
        "entrambi i contesti? Rispondere con {labels_str}, e nient'altro.",
    ),
    LATVIAN: PromptConfig(
        default_prompt_label_mapping=dict(same_sense="jā", different_sense="nē"),
        default_prompt_prefix="Turpmāk ir piemēri vārdiem, kas izmantoti divos "
        "kontekstos, un vai tiem ir vienāda nozīme.",
        default_prompt_template="{text}\nVienāda nozīme: {label}",
        default_instruction_prompt="{text}\n\nVai vārdam ir vienāda nozīme abos "
        "kontekstos? Atbildiet ar {labels_str}, un neko citu.",
    ),
    LITHUANIAN: PromptConfig(
        default_prompt_label_mapping=dict(same_sense="taip", different_sense="ne"),
        default_prompt_prefix="Toliau pateikti pavyzdžiai, kaip žodžiai vartojami "
        "dviejuose kontekstuose, ir ar jie turi tą pačią reikšmę.",
        default_prompt_template="{text}\nTa pati reikšmė: {label}",
        default_instruction_prompt="{text}\n\nAr žodis turi tą pačią reikšmę "
        "abiejuose kontekstuose? Atsakykite su {labels_str}, ir nieko kito.",
    ),
    NORWEGIAN_BOKMÅL: PromptConfig(
        default_prompt_label_mapping=dict(same_sense="ja", different_sense="nei"),
        default_prompt_prefix="Følgende er eksempler på ord brukt i to kontekster og "
        "om de har samme betydning.",
        default_prompt_template="{text}\nSamme betydning: {label}",
        default_instruction_prompt="{text}\n\nHar ordet samme betydning i begge "
        "kontekstene? Svar med {labels_str}, og ikke noe annet.",
    ),
    NORWEGIAN_NYNORSK: PromptConfig(
        default_prompt_label_mapping=dict(same_sense="ja", different_sense="nei"),
        default_prompt_prefix="Følgjande er eksempel på ord brukte i to kontekstar og "
        "om dei har same tyding.",
        default_prompt_template="{text}\nSame tyding: {label}",
        default_instruction_prompt="{text}\n\nHar ordet same tyding i begge "
        "kontekstane? Svar med {labels_str}, og ikkje noko anna.",
    ),
    NORWEGIAN: PromptConfig(
        default_prompt_label_mapping=dict(same_sense="ja", different_sense="nei"),
        default_prompt_prefix="Følgende er eksempler på ord brukt i to kontekster og "
        "om de har samme betydning.",
        default_prompt_template="{text}\nSamme betydning: {label}",
        default_instruction_prompt="{text}\n\nHar ordet samme betydning i begge "
        "kontekstene? Svar med {labels_str}, og ikke noe annet.",
    ),
    POLISH: PromptConfig(
        default_prompt_label_mapping=dict(same_sense="tak", different_sense="nie"),
        default_prompt_prefix="Poniżej znajdują się przykłady słów użytych w dwóch "
        "kontekstach i czy mają to samo znaczenie.",
        default_prompt_template="{text}\nTo samo znaczenie: {label}",
        default_instruction_prompt="{text}\n\nCzy słowo ma to samo znaczenie w obu "
        "kontekstach? Odpowiedz używając wyłącznie {labels_str}.",
    ),
    PORTUGUESE: PromptConfig(
        default_prompt_label_mapping=dict(same_sense="sim", different_sense="não"),
        default_prompt_prefix="A seguir estão exemplos de palavras usadas em dois "
        "contextos e se têm o mesmo significado.",
        default_prompt_template="{text}\nMesmo significado: {label}",
        default_instruction_prompt="{text}\n\nA palavra tem o mesmo significado em "
        "ambos os contextos? Responde com {labels_str}, e nada mais.",
    ),
    ROMANIAN: PromptConfig(
        default_prompt_label_mapping=dict(same_sense="da", different_sense="nu"),
        default_prompt_prefix="Următoarele sunt exemple de cuvinte folosite în două "
        "contexte și dacă au același sens.",
        default_prompt_template="{text}\nAcelași sens: {label}",
        default_instruction_prompt="{text}\n\nAre cuvântul același sens în ambele "
        "contexte? Răspundeți cu {labels_str}, și nimic altceva.",
    ),
    SERBIAN: PromptConfig(
        default_prompt_label_mapping=dict(same_sense="da", different_sense="ne"),
        default_prompt_prefix="U nastavku su primjeri reči korišćenih u dva konteksta "
        "i da li imaju isto značenje.",
        default_prompt_template="{text}\nIsto značenje: {label}",
        default_instruction_prompt="{text}\n\nIma li reč isto značenje u oba "
        "konteksta? Odgovorite sa {labels_str}, i ništa drugo.",
    ),
    SLOVAK: PromptConfig(
        default_prompt_label_mapping=dict(same_sense="áno", different_sense="nie"),
        default_prompt_prefix="Nasledujú príklady slov použitých v dvoch kontextoch "
        "a či majú rovnaký význam.",
        default_prompt_template="{text}\nRovnaký význam: {label}",
        default_instruction_prompt="{text}\n\nMá slovo v oboch kontextoch rovnaký "
        "význam? Odpovedzte so {labels_str}, a nič iné.",
    ),
    SLOVENE: PromptConfig(
        default_prompt_label_mapping=dict(same_sense="da", different_sense="ne"),
        default_prompt_prefix="Sledeči so primeri besed, ki se uporabljajo v dveh "
        "kontekstih, in ali imajo enak pomen.",
        default_prompt_template="{text}\nEnak pomen: {label}",
        default_instruction_prompt="{text}\n\nAli ima beseda enak pomen v obeh "
        "kontekstih? Odgovorite z {labels_str}, in nič drugega.",
    ),
    SPANISH: PromptConfig(
        default_prompt_label_mapping=dict(same_sense="sí", different_sense="no"),
        default_prompt_prefix="A continuación se presentan ejemplos de palabras usadas "
        "en dos contextos y si tienen el mismo significado.",
        default_prompt_template="{text}\nMismo significado: {label}",
        default_instruction_prompt="{text}\n\n¿Tiene la palabra el mismo significado "
        "en ambos contextos? Responde con {labels_str}, y nada más.",
    ),
    SWEDISH: PromptConfig(
        default_prompt_label_mapping=dict(same_sense="ja", different_sense="nej"),
        default_prompt_prefix="Följande är exempel på ord använda i två sammanhang och "
        "om de har samma betydelse.",
        default_prompt_template="{text}\nSamma betydelse: {label}",
        default_instruction_prompt="{text}\n\nHar ordet samma betydelse i båda "
        "sammanhangen? Svara med {labels_str}, och inget annat.",
    ),
    UKRAINIAN: PromptConfig(
        default_prompt_label_mapping=dict(same_sense="так", different_sense="ні"),
        default_prompt_prefix="Нижче наведені приклади слів, що використовуються в "
        "двох контекстах, і чи мають вони однакове значення.",
        default_prompt_template="{text}\nОднакове значення: {label}",
        default_instruction_prompt="{text}\n\nЧи має слово однакове значення в обох "
        "контекстах? Відповідайте {labels_str}, і нічого більше.",
    ),
}



# /// script
# requires-python = ">=3.10,<4.0"
# dependencies = [
#     "pandas==2.2.0",
#     "requests==2.32.3",
# ]
# ///

"""Load the part-of-speech part of a Universal Dependencies treebank."""

import collections.abc as c
import logging
import re
import typing as t
from collections import defaultdict
from functools import partial
from urllib.parse import urlparse

import pandas as pd
import requests

logging.basicConfig(format="%(asctime)s ⋅ %(message)s", level=logging.INFO)
logger = logging.getLogger("load_ud_pos")


def load_dadt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Danish Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val` and `test`.
    """
    # Define download URLs
    base_url = (
        "https://github.com/UniversalDependencies/UD_Danish-DDT/raw/master/"
        "da_ddt-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_ptdt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Portuguese Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val` and `test`.
    """
    # Define download URLs
    base_url = (
        "https://github.com/UniversalDependencies/UD_Portuguese-Bosque/raw/master/"
        "pt_bosque-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(
        train_url=train_url,
        val_url=val_url,
        test_url=test_url,
        filter_source="CETEMPúblico",
    )


def load_fodt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Faroese Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val` and `test`.
    """
    # Define download URLs
    base_url = (
        "https://github.com/UniversalDependencies/UD_Faroese-FarPaHC/raw/master/"
        "fo_farpahc-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_isdt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Icelandic Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val` and `test`.
    """
    # Define download URLs
    base_url = (
        "https://github.com/UniversalDependencies/UD_Icelandic-Modern/raw/master/"
        "is_modern-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_nodt_nb_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Norwegian Bokmål Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val` and `test`.
    """
    # Define download URLs
    base_url = (
        "https://github.com/UniversalDependencies/UD_Norwegian-Bokmaal/raw/master/"
        "no_bokmaal-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_nodt_nn_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Norwegian Nynorsk Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val` and `test`.
    """
    # Define download URLs
    base_url = (
        "https://github.com/UniversalDependencies/UD_Norwegian-Nynorsk/raw/master/"
        "no_nynorsk-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_svdt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Swedish Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val` and `test`.
    """
    # Define download URLs
    base_url = (
        "https://github.com/UniversalDependencies/UD_Swedish-Talbanken/raw/master/"
        "sv_talbanken-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    # Define document processing function
    def process_document(doc: str) -> str:
        doc = (
            doc.replace(" s k", " s.k.")
            .replace("S k", "S.k.")
            .replace(" bl a", " bl.a.")
            .replace("Bl a", "Bl.a.")
            .replace(" t o m", " t.o.m.")
            .replace("T o m", "T.o.m.")
            .replace(" fr o m", " fr.o.m.")
            .replace("Fr o m", "Fr.o.m.")
            .replace(" o s v", " o.s.v.")
            .replace("O s v", "O.s.v.")
            .replace(" d v s", " d.v.s.")
            .replace("D v s", "D.v.s.")
            .replace(" m fl", " m.fl.")
            .replace("M fl", "M.fl.")
            .replace(" t ex", " t.ex.")
            .replace("T ex", "T.ex.")
            .replace(" f n", " f.n.")
            .replace("F n", "F.n.")
        )
        return doc

    return load_ud_pos(
        train_url=train_url,
        val_url=val_url,
        test_url=test_url,
        doc_process_fn=process_document,
    )


def load_dedt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the German Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val` and `test`.
    """
    # Define download URLs
    base_url = (
        "https://raw.githubusercontent.com/UniversalDependencies/UD_German-GSD/master/"
        "de_gsd-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_nldt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Dutch Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val` and `test`.
    """
    # Define download URLs
    base_url = (
        "https://raw.githubusercontent.com/UniversalDependencies/UD_Dutch-Alpino/"
        "master/nl_alpino-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_endt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the English Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val` and `test`.
    """
    # Define download URLs
    base_url = (
        "https://raw.githubusercontent.com/UniversalDependencies/UD_English-GUM/"
        "master/en_gum-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_frdt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the French Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val` and `test`.
    """
    # Define download URLs
    base_url = (
        "https://raw.githubusercontent.com/UniversalDependencies/UD_French-GSD/"
        "master/fr_gsd-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_itdt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Italian Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val` and `test`.
    """
    # Define download URLs
    base_url = (
        "https://raw.githubusercontent.com/UniversalDependencies/UD_Italian-ISDT/"
        "master/it_isdt-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_esdt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Spanish Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val` and `test`.
    """
    # Define download URLs
    base_url = (
        "https://raw.githubusercontent.com/UniversalDependencies/UD_Spanish-AnCora"
        "/refs/heads/master/es_ancora-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_fidt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Finnish Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val` and `test`.
    """
    # Define download URLs
    base_url = (
        "https://raw.githubusercontent.com/UniversalDependencies/UD_Finnish-TDT/"
        "master/fi_tdt-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_lvdt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Latvian Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val` and `test`.
    """
    # Define download URLs
    base_url = (
        "https://raw.githubusercontent.com/UniversalDependencies/UD_Latvian-LVTB/"
        "master/lv_lvtb-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_etdt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Estonian Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val` and `test`.
    """
    # Define download URLs
    base_url = (
        "https://raw.githubusercontent.com/UniversalDependencies/UD_Estonian-EDT/"
        "refs/heads/master/et_edt-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_pldt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Polish Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val` and `test`.
    """
    # Define download URLs
    base_url = (
        "https://raw.githubusercontent.com/UniversalDependencies/UD_Polish-PDB/"
        "refs/heads/master/pl_pdb-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_ltdt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Lithuanian Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val` and `test`.
    """
    # Define download URLs
    base_url = (
        "https://raw.githubusercontent.com/UniversalDependencies/UD_Lithuanian-ALKSNIS/"
        "master/lt_alksnis-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_csdt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Czech Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val` and `test`.
    """
    # Define download URLs
    base_url = (
        "https://raw.githubusercontent.com/UniversalDependencies/UD_Czech-CAC/refs/"
        "heads/master/cs_cac-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_skdt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Slovak Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val`, and `test`.
    """
    # Define download URLs
    base_url = (
        "https://raw.githubusercontent.com/UniversalDependencies/UD_Slovak-SNK/refs"
        "/heads/master/sk_snk-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_ukdt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Ukrainian Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val` and `test`.
    """
    # Define download URLs
    base_url = (
        "https://raw.githubusercontent.com/UniversalDependencies/"
        "UD_Ukrainian-ParlaMint/refs/heads/master/uk_parlamint-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_eldt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Greek Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val`, and `test`.
    """
    # Define download URLs
    base_url = (
        "https://raw.githubusercontent.com/UniversalDependencies/UD_Greek-GUD/refs"
        "/heads/master/el_gud-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_bgdt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Bulgarian Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val` and `test`.
    """
    # Define download URLs
    base_url = (
        "https://raw.githubusercontent.com/UniversalDependencies/UD_Bulgarian-BTB/refs"
        "/heads/master/bg_btb-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_srdt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Serbian Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val`, and `test`.
    """
    # Define download URLs
    base_url = (
        "https://raw.githubusercontent.com/UniversalDependencies/UD_Serbian-SET/refs"
        "/heads/master/sr_set-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_sldt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Slovene Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val`, and `test`.
    """
    # Define download URLs
    base_url = (
        "https://raw.githubusercontent.com/UniversalDependencies/UD_Slovene-SSJ/refs"
        "/heads/master/sl_ssj-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_hrdt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Croatian Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val`, and `test`.
    """
    # Define download URLs
    base_url = (
        "https://raw.githubusercontent.com/UniversalDependencies/UD_Croatian-SET/refs"
        "/heads/master/hr_set-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_hudt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Hungarian Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val`, and `test`.
    """
    # Define download URLs
    base_url = (
        "https://raw.githubusercontent.com/UniversalDependencies/UD_Hungarian-Szeged/refs"
        "/heads/master/hu_szeged-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_rodt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Romanian Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val`, and `test`.
    """
    # Define download URLs
    base_url = (
        "https://raw.githubusercontent.com/UniversalDependencies/UD_Romanian-RRT/refs"
        "/heads/master/ro_rrt-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_cadt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Catalan Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val`, and `test`.
    """
    # Define download URLs
    base_url = (
        "https://raw.githubusercontent.com/UniversalDependencies/UD_Catalan-AnCora/refs"
        "/heads/master/ca_ancora-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def load_sqdt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Albanian Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val`, and `test`.
    """
    # Define download URLs
    base_url = (
        "https://raw.githubusercontent.com/UniversalDependencies/UD_Albanian-STAF/refs"
        "/heads/master/sq_staf-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    df_dict = load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)

    # Load additional TSA test data and concatenate with the existing test split
    tsa_test_url = (
        "https://raw.githubusercontent.com/UniversalDependencies/UD_Albanian-TSA/refs"
        "/heads/master/sq_tsa-ud-test.conllu"
    )
    tsa_test_lines = _load_file_or_url(url_or_path=tsa_test_url)
    tsa_test_df = _load_split(lines=tsa_test_lines)

    df_dict["test"] = pd.concat([df_dict["test"], tsa_test_df], ignore_index=True)

    return df_dict


def load_bedt_pos() -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of the Belarusian Dependency Treebank.

    Returns:
        The dataframes, stored in the keys `train`, `val`, and `test`.
    """
    # Define download URLs
    base_url = (
        "https://raw.githubusercontent.com/UniversalDependencies/UD_Belarusian-HSE/refs"
        "/heads/master/be_hse-ud-{}.conllu"
    )
    train_url = base_url.format("train")
    val_url = base_url.format("dev")
    test_url = base_url.format("test")

    return load_ud_pos(train_url=train_url, val_url=val_url, test_url=test_url)


def _load_file_or_url(url_or_path: str) -> list[str]:
    """Load a file from a URL or local path.

    Args:
        url_or_path: The URL or local path to load.

    Returns:
        The list of strings, one per line.
    """
    parsed = urlparse(url_or_path)
    if parsed.scheme.lower() in ("http", "https"):
        return requests.get(url_or_path).text.split("\n")
    else:
        with open(url_or_path, "r") as f:
            logger.warning(f"Loading data from local file: {url_or_path}")
            return [line.strip() for line in f.readlines()]


def _filter_token_range(data_dict: dict[str, list]) -> dict[str, list]:
    """Filter out tokens that belong to ranges in UD source files.

    Tokens that span more than one position are not supported by create_scala's
    prepare_df logic.

    Example files:

    - tests/test_scripts/test_create_scala/test_data/de_gsd-ud-train.conllu.adp_det
    - tests/test_scripts/test_create_scala/test_data/pl_pdb-ud-train.conllu.aux_clitic_*

    Args:
        data_dict: The input data dictionary.

    Returns:
        The filtered data dictionary. Its format is identical to the input.
    """
    output: t.DefaultDict[str, list] = defaultdict(list)

    range_start: int = 0
    range_end: int = 0
    # Track the index where we added the multi-word token so we can update its POS tag
    mwt_index: int | None = None

    # Pattern used to detect merged IDs
    merged_ids_pattern = re.compile(r"(\d+)-(\d+)", re.I)

    for i in range(len(data_dict["ids"])):
        match = merged_ids_pattern.match(data_dict["ids"][i])
        if match is not None:
            output["ids"].append(data_dict["ids"][i])
            output["tokens"].append(data_dict["tokens"][i])
            output["pos_tags"].append(data_dict["pos_tags"][i])
            range_start = int(match.group(1))
            range_end = int(match.group(2))
            # Remember where we added this multi-word token
            mwt_index = len(output["pos_tags"]) - 1
        else:
            token_id = int(data_dict["ids"][i].split(".")[0])
            if token_id >= range_start and token_id <= range_end:
                # If this is the first token in the range and we have a multi-word token
                # to update, copy the POS tag from this token to the multi-word token
                if mwt_index is not None and token_id == range_start:
                    output["pos_tags"][mwt_index] = data_dict["pos_tags"][i]
                    mwt_index = None  # Reset so we only update once per range
                # Skip token if in range
                continue
            else:
                output["ids"].append(data_dict["ids"][i])
                output["tokens"].append(data_dict["tokens"][i])
                output["pos_tags"].append(data_dict["pos_tags"][i])
                # Reset mwt_index when we move past the range
                mwt_index = None

    return output


def _load_split(
    *,
    lines: list[str],
    filter_source: str | None = None,
    doc_process_fn: c.Callable[[str], str] = lambda x: x,
) -> pd.DataFrame:
    """Load single split of the POS part of a Universal Dependencies treebank.

    Args:
        lines: The lines of the file to process.
        filter_source:
            If not `None`, only include entries with this source in the dataset.
        doc_process_fn:
            A function to apply to each document before parsing it.

    Returns:
        The dataframe for the given split.
    """
    # Initialise the records, data dictionary and document
    records: list[dict] = []
    data_dict: dict[str, list[list[int | str] | str]] = defaultdict(list)

    # Iterate over the data for the given split
    doc = ""
    source = ""
    for line_idx, line in enumerate(lines):
        # If we are at the first line of an entry then extract the document
        if line.startswith("# text = "):
            doc = re.sub("# text = ", "", line)

            # Process the document if needed
            doc = doc_process_fn(doc)

        elif line.startswith("# source = "):
            source = line.removeprefix("# source = ").strip()

        # Otherwise, if the line is a comment then ignore it
        elif line.startswith("#"):
            continue

        # Otherwise, if we have reached the end of an entry then store it to the
        # list of records and reset the data dictionary and document
        elif line == "" or line_idx == len(lines) - 1:
            if len(data_dict["tokens"]) > 0:
                if filter_source is None or filter_source in source:
                    merged_data_dict: dict[str, str | list[int | str]]
                    merged_data_dict = {**_filter_token_range(data_dict), "doc": doc}
                    records.append(merged_data_dict)
            data_dict = defaultdict(list)
            doc = ""
            source = ""

        # Otherwise we are in the middle of an entry which is not a comment, so
        # we extract the data from the line and store it in the data dictionary
        else:
            data_tup = line.split("\t")
            data_dict["ids"].append(data_tup[0])
            data_dict["tokens"].append(data_tup[1])
            data_dict["pos_tags"].append(data_tup[3])

    # Convert the records to a dataframe
    return pd.DataFrame.from_records(records)


def load_ud_pos(
    train_url: str,
    val_url: str,
    test_url: str,
    doc_process_fn: c.Callable[[str], str] = lambda x: x,
    filter_source: str | None = None,
) -> dict[str, pd.DataFrame]:
    """Load the part-of-speech part of a Universal Dependencies treebank.

    Args:
        train_url:
            The URL of the training data.
        val_url:
            The URL of the validation data.
        test_url:
            The URL of the test data.
        doc_process_fn:
            A function to apply to each document before parsing it.
        filter_source:
            If not `None`, only include entries with this source in the dataset.

    Returns:
        The dataframes, stored in the keys `train`, `val` and `test`.
    """
    if filter_source is not None:
        logger.warning(
            f"Warning: Filtering dataset to include only entries with {filter_source=}"
        )

    # Load the lines of the splits
    train_lines = _load_file_or_url(url_or_path=train_url)
    val_lines = _load_file_or_url(url_or_path=val_url)
    test_lines = _load_file_or_url(url_or_path=test_url)

    # Load the splits
    split_loader = partial(
        _load_split, filter_source=filter_source, doc_process_fn=doc_process_fn
    )
    dfs = dict(
        train=split_loader(lines=train_lines),
        val=split_loader(lines=val_lines),
        test=split_loader(lines=test_lines),
    )

    # Return the dictionary of dataframes
    return dfs



# /// script
# requires-python = ">=3.10,<4.0"
# dependencies = [
#     "datasets==3.5.0",
#     "huggingface-hub==0.24.0",
#     "pandas==2.2.0",
#     "requests==2.32.3",
# ]
# ///

"""Create the Ukrainian NER dataset and upload it to the HF Hub."""

import pandas as pd
from datasets import Dataset, DatasetDict, Split, load_dataset
from huggingface_hub import HfApi


def main() -> None:
    """Create the Ukrainian NER dataset and upload it to the HF Hub."""
    # Define dataset ID
    repo_id = "benjamin/ner-uk"

    # Load the dataset
    dataset = load_dataset(path=repo_id)
    assert isinstance(dataset, DatasetDict)

    # Convert the dataset to dataframes
    train_df = dataset["train"].to_pandas()
    val_df = dataset["validation"].to_pandas()
    test_df = dataset["test"].to_pandas()

    assert isinstance(train_df, pd.DataFrame)
    assert isinstance(val_df, pd.DataFrame)
    assert isinstance(test_df, pd.DataFrame)

    # Add a `text` column by joining tokens
    train_df["text"] = train_df["tokens"].map(lambda tokens: " ".join(tokens))
    val_df["text"] = val_df["tokens"].map(lambda tokens: " ".join(tokens))
    test_df["text"] = test_df["tokens"].map(lambda tokens: " ".join(tokens))

    # Rename `ner_tags` to `labels`
    train_df.rename(columns={"ner_tags": "labels"}, inplace=True)
    val_df.rename(columns={"ner_tags": "labels"}, inplace=True)
    test_df.rename(columns={"ner_tags": "labels"}, inplace=True)

    ner_conversion_dict = {
        0: "O",
        1: "B-PER",
        2: "I-PER",
        3: "B-ORG",
        4: "I-ORG",
        5: "B-LOC",
        6: "I-LOC",
        7: "B-MISC",
        8: "I-MISC",
    }

    train_df["labels"] = train_df["labels"].map(
        lambda ner_tags: [ner_conversion_dict[ner_tag] for ner_tag in ner_tags]
    )
    val_df["labels"] = val_df["labels"].map(
        lambda ner_tags: [ner_conversion_dict[ner_tag] for ner_tag in ner_tags]
    )
    test_df["labels"] = test_df["labels"].map(
        lambda ner_tags: [ner_conversion_dict[ner_tag] for ner_tag in ner_tags]
    )

    column_to_keep = ["tokens", "labels"]
    train_df = train_df[column_to_keep]
    val_df = val_df[column_to_keep]
    test_df = test_df[column_to_keep]

    # Make splits
    train_size = 1024
    val_size = 256
    test_size = 2048

    # Calculate the number of samples to transfer from train to test
    current_test_size = len(test_df)
    extra_samples_needed = test_size - current_test_size

    # Sample additional from the train set
    extra_test_samples = train_df.sample(n=extra_samples_needed, random_state=4242)

    # Add these samples to the test set and remove them from the train set
    test_df = pd.concat([test_df, extra_test_samples]).reset_index(drop=True)
    train_df = train_df.drop(extra_test_samples.index.tolist()).reset_index(drop=True)

    # Sample the desired sizes
    train_df = train_df.sample(n=train_size, random_state=4242)
    val_df = val_df.sample(n=val_size, random_state=4242)

    # Confirm the test size
    assert len(test_df) == test_size

    # Reset the index
    train_df.reset_index(drop=True, inplace=True)
    val_df.reset_index(drop=True, inplace=True)
    test_df.reset_index(drop=True, inplace=True)

    assert isinstance(train_df, pd.DataFrame)
    assert isinstance(val_df, pd.DataFrame)
    assert isinstance(test_df, pd.DataFrame)

    # Create the dataset dictionary after resetting the index
    dataset = DatasetDict(
        {
            "train": Dataset.from_pandas(train_df, split=Split.TRAIN),
            "val": Dataset.from_pandas(val_df, split=Split.VALIDATION),
            "test": Dataset.from_pandas(test_df, split=Split.TEST),
        }
    )

    # Create dataset ID
    dataset_id = "EuroEval/ner-uk-mini"

    # Remove the dataset from Hugging Face Hub if it already exists
    HfApi().delete_repo(dataset_id, repo_type="dataset", missing_ok=True)

    # Push the dataset to the Hugging Face Hub
    dataset.push_to_hub(dataset_id, private=True)


if __name__ == "__main__":
    main()
